<!DOCTYPE html>
<html lang="en">

<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <link rel="stylesheet" href="{{ url_for('static', filename='./styles/start.css') }}">
    <link rel="stylesheet" href="{{ url_for('static', filename='./styles/input.css') }}">
    <link rel="stylesheet" href="{{ url_for('static', filename='./styles/microsoft.css') }}">
    <link rel="shortcut icon" href="{{ url_for('static', filename='./assets/logo.png') }}" type="image/x-icon">
    <script src="https://ajax.googleapis.com/ajax/libs/jquery/3.2.1/jquery.min.js"></script>
    <title>SIENA - error</title>
    
</head>

<body>
    <h1>Something went wrong</h1>
</body>
</html>